#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
*****************************************************************************
LIRIS action annotation tool

Authors: Christian Wolf, Eric Lombardi, Julien Mille

*****************************************************************************
"""

from Tkinter import Tk, Canvas, Frame, BOTH, Listbox, Toplevel, Message, Button, Entry
from Tkinter import N,S,W,E,NW, SW, NE, SE, CENTER, END, LEFT, RIGHT, X, Y, TOP, BOTTOM
import Image
import ImageTk
import sys, glob, copy
import tkMessageBox
import os
import xml.etree.ElementTree as xml


# ***************************************************************************
# Global constants
# ***************************************************************************

MAX_RUNID = 100
CORNER_DIST_THR=30
CENTER_DIST_THR=50
CORNER_SIZE=30
CENTER_SIZE=30
JUMP_FRAMES=25
TITLE = "LIRIS-HARL action reader V1.01"

# ***************************************************************************
# The data structure storing the annotations
# ***************************************************************************

classnames = ["discussion","give","box/desk","enter/leave","try-to-enter","unlock+enter","baggage","handshaking","typing","telephone"]

# ***************************************************************************
# XML parsing helper functions
# ***************************************************************************

# Return the the given tag in the given tree, and ensure that this tag is 
# present only a single time. 
def getSingleTag(tree,tagname):
    rv=tree.findall(tagname)
    if len(rv)!=1:
        tkMessageBox.showinfo(TITLE, "tag "+tagname+" needs to occur a single time at this point!")
        sys.exit(1)
    return rv[0] 

# Return an attribute value. Check for its existence
def getAtt(node,attname):
    rv=node.get(attname)
    if rv==None:
        tkMessageBox.showinfo(TITLE, "attribute "+attname+" not found in tag "+node.tag)
        sys.exit(1)
    return rv

# ***************************************************************************
class AARect:
    """A rectangle (bounding box) and its running id"""
    def __init__(self,x1,y1,x2,y2,runid):
        if x1<x2:
            self.x1=x1
            self.x2=x2
        else:
            self.x1=x2
            self.x2=x1
        if y1<y2:
            self.y1=y1	
            self.y2=y2
        else:
            self.y1=y2	
            self.y2=y1
        self.runid=runid
        
    def show(self):
        print "x1=", self.x1, "  y1=", self.y1, "  x2=", self.x2, "  y2=", self.y2, "  id=", self.runid



# ***************************************************************************
class SemMousePos:
    """A semantic mouse position: in which rectangle (index) is the mouse
       and which semantic position does it occupy.
       sempose can be:
       ul	upper left corner
       ur	upper right corner
       ll	lower left corner
       lr	lower right corner
       c	center
       g	general position in the recangle
       n	no rectangles"""
    def __init__(self,index,sempos):
        self.index=index
        self.sempos=sempos          

# ***************************************************************************
class AAFrame:
    """All rectangles of a frame"""
    def __init__(self):
        self.rects=[]
	
    def getRects(self):
        return self.rects
	
    # Check the position of the mouse cursor with respect to corners of all
    # the rectangles, as well as the centers. If it is not near anything, 
    # still check for the nearest center.
    # position x,y
    def getSemMousePos(self,x,y):
	
        # First check for the corners
        minval=99999999
        argindex=-1	
        for (i,r) in enumerate(self.rects):	    
            d=(r.x1-x)*(r.x1-x)+(r.y1-y)*(r.y1-y)
            if d<minval:
                minval=d
                argindex=i
                argsem="ul"
            d=(r.x1-x)*(r.x1-x)+(r.y2-y)*(r.y2-y)
            if d<minval:
                minval=d
                argindex=i
                argsem="ll"
            d=(r.x2-x)*(r.x2-x)+(r.y1-y)*(r.y1-y)
            if d<minval:
                minval=d
                argindex=i
                argsem="ur"
            d=(r.x2-x)*(r.x2-x)+(r.y2-y)*(r.y2-y)
            if d<minval:
                minval=d
                argindex=i
                argsem="lr"
            
        # We near enough to a corner, we are done
        if minval<CORNER_DIST_THR*CORNER_DIST_THR:
            return SemMousePos(argindex,argsem)
            
        # Now check for the nearest center
        minval=99999999
        argindex=-1
        for (i,r) in enumerate(self.rects):
            cx=0.5*(r.x1+r.x2)
            cy=0.5*(r.y1+r.y2)
            d=(cx-x)*(cx-x)+(cy-y)*(cy-y)
            if d<minval:
                minval=d
                argindex=i
        
        if argindex<0:
            return SemMousePos(-1,"n");
        
        if minval<CENTER_DIST_THR*CENTER_DIST_THR:
            return SemMousePos(argindex,"c")
        else:
            return SemMousePos(argindex,"g")	

# ***************************************************************************
class AAControler:            
    def __init__(self):
        # An array holding an AAFrame object for each frame of the video
        self.frames=[]	
        # An array holding the classnr for each action nr. ("runid")
        self.runids=[]
        # The nr. of the currently visible frame
        self.curFrameNr=0
        self.videoname=""
        if len(sys.argv)!=3:
            self.usage()
        prefix=sys.argv[1]
        self.outputfilename=sys.argv[2]
        self.filenames=sorted(glob.glob(prefix+"*"))
        # print self.filenames
        if len(self.filenames)<1:
            print >> sys.stderr, "Did not find any frames! Is the prefix correct?"
            self.usage()
        for i in range(len(self.filenames)):
            self.frames.append(AAFrame())
            
        # If the given XML file exists, parse it
        if os.path.isfile(self.outputfilename):
            self.parseXML()
        else:
        
            # If it does NOT exist, let's try to create one
            try:
                fd=open(self.outputfilename,'w')
                
            # Unsuccessful -> the given directory does not exist
            except:
                s="Could not save to the specified XML file. Please check the location. Does the directory exist?"
                tkMessageBox.showinfo(TITLE,s)
                sys.exit(1)
            tkMessageBox.showinfo(TITLE, "XML File "+self.outputfilename+" does not exist. Creating a new one.")
        
    def usage(self):
        print >> sys.stderr, "usage:"
        print >> sys.stderr, sys.argv[0],"<framefileprefix> <output-xml-filename>"
        sys.exit(1);
    
    # Check the current annotation for validity
    def checkValidity(self):
        msg=''
    
        # Check for non contiguous activities.
        # Keep a dictionary which holds for each activity the framenr of the
        # last frame
        acts = {}
        for (frnr,fr) in enumerate(self.frames):
            for r in fr.rects:
                if r.runid in acts:
                    if frnr-acts[r.runid] > 1:
                        msg = msg+"Activity nr. "+str(r.runid)+" has a hole after frame nr. "+str(acts[r.runid])+".\n"
                acts[r.runid] = frnr;
                    
        # Check for several occurrences of a runid in the same frame.
        for (frnr,fr) in enumerate(self.frames):
            msg=msg+self.checkValidityFrame(frnr)
        
        # Check for unassigned runids (no known action class)
        msg2=''
        for (i,x) in enumerate(self.runids):
            if x<0:
                msg2=msg2+str(i+1)+","
        if msg2<>'':
            msg2="The following activities do not have assigned classes: "+msg2+"\n"
        msg=msg+msg2    
            
        return msg 
        
    # Check a single frame for validity (multiple identical runids)    
    def checkValidityFrame(self, framenr):
        msg=''
        ids=set()
        for r in self.frames[framenr].rects:
            if r.runid in ids:
                msg = msg+'Activity nr. '+str(r.runid)+' occurs multiple times in frame nr. '+str(framenr)+'.\n'
            else:
                ids.add(r.runid)
        return msg

    # Open the image corresponding to the current frame number,
    # set the property self.curImage, and return it
    def curFrame(self):
        self.curImage = Image.open(self.filenames[self.curFrameNr])
        # print "frame nr. ",self.curFrameNr, "=",self.filenames[self.curFrameNr]
        return self.curImage
    
    # Remove all rectangles of the current frame
    def deleteAllRects(self):
        self.frames[self.curFrameNr].rects = []
      
    # Remove the rectangle with the given index from the list
    # of rectangles of the currently selected frame
    def deleteRect(self, index):
        del self.frames[self.curFrameNr].rects[index];
        
    def nextFrame(self,doPropagate,force):                    
        if self.curFrameNr<len(self.filenames)-1:
            self.curFrameNr+=1        
        # if the next frame does NOT contain any rectangles, 
        # propagate the previous ones
        if doPropagate:
            x=len(self.frames[self.curFrameNr].rects)
        else:
            self.curFrame()
        self.exportXMLFilename("save.xml")
        return self.curImage
        
    def nextFrameFar(self):
        if self.curFrameNr<len(self.filenames)-JUMP_FRAMES:
            self.curFrameNr+=JUMP_FRAMES 
        else:
            self.curFrameNr=len(self.filenames)-1
        self.exportXMLFilename("save.xml")
        return self.curFrame()
	
    def prevFrame(self):
        if self.curFrameNr>0:
            self.curFrameNr-=1
        self.exportXMLFilename("save.xml")
        return self.curFrame()  
        
    def prevFrameFar(self):
        if self.curFrameNr>=JUMP_FRAMES:
            self.curFrameNr-=JUMP_FRAMES
        else:
            self.curFrameNr=0
        self.exportXMLFilename("save.xml")
        return self.curFrame()  
                
    def getRects(self):	
        return self.frames[self.curFrameNr].getRects()
	
    def addRect(self,x1,y1,x2,y2,runid,fnr=-1):
        if fnr==-1:
            fnr=self.curFrameNr
        if fnr>=len(self.frames):
            raise Exception()
        self.frames[fnr].getRects().append(AARect(x1,y1,x2,y2,runid))
	
    def delRect(self,index):	
        del self.frames[self.curFrameNr].getRects()[index]
	
    def getSemMousePos(self,x,y):
        return self.frames[self.curFrameNr].getSemMousePos(x,y)
	    			
	# Update the running id for a rectangle index
    def updateRunId(self,indexRect,newId):        
        self.frames[self.curFrameNr].rects[indexRect].runid=newId	
        self.useRunId(newId)
        
    # Tell the system the given runId is used. If the array holding the classes 
    # for the different ids is not large enough, grow it and insert -1 as class
    def useRunId(self,newId):
        neededcap=newId-len(self.runids)
        if neededcap>0:
            for i in range(neededcap):
                self.runids.append(-1)
        print "new run id array",self.runids
        
    def exportXML(self):
        self.exportXMLFilename(self.outputfilename)
	
    def exportXMLFilename(self,filename):
        # Get maximum running id 
        maxid=-1
        for (i,f) in enumerate(self.frames):
            for (j,r) in enumerate(f.getRects()):
                if r.runid > maxid:
                    maxid=r.runid
        
        try:
            fd=open(filename,'w')
        except:
            tkMessageBox.showinfo(TITLE, "Could not save to the specified XML file. Please check the location. Does the directory exist?")
        print >> fd, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
        print >> fd, "<tagset>"
        print >> fd, "  <video>"
        print >> fd, "    <videoName>"+self.videoname+"</videoName>"
        
        # Travers all different running id's 
        for currunid in range(maxid):
            foundRects=False
            for (i,f) in enumerate(self.frames):
                for (j,r) in enumerate(f.getRects()):
                    if r.runid==currunid+1:
                        if not foundRects:
                            foundRects=True
                            fd.write ("    <action nr=\""+str(currunid+1)+"\" class=\""+str(self.runids[currunid])+"\">\n")	
                        s="      <bbox x=\""+str(int(r.x1))+"\" y=\""+str(int(r.y1))
                        s=s+"\" width=\""+str(int(r.x2-r.x1+1))+"\" height=\""+str(int(r.y2-r.y1+1))
                        s=s+"\" framenr=\""+str(i+1)+"\"/>\n"
                        fd.write(s)
            if foundRects:
                print >> fd, "    </action>"	    	    
        print >> fd, "  </video>"  
        print >> fd, "</tagset>"  
        fd.close()
        
    def parseXML(self):
        tree = xml.parse(self.outputfilename)
        rootElement = tree.getroot()
        
        # Get the single video tag
        vids=tree.findall("video")
        if len(vids)<1:
            tkMessageBox.showinfo(TITLE, "No <video> tag found in the input XML file!")
            sys.exit(1)
        if len(vids)>1:
            tkMessageBox.showinfo(TITLE, "Currently only a single <video> tag is supported per XML file!")
            sys.exit(1)
        vid=vids[0]
        
        # Get the video name
        x=getSingleTag(vid,"videoName")            
        if (x.text is None) or (len(x.text)==0) or (x.text=="NO-NAME"):
            tkMessageBox.showinfo(TITLE, "The video name in the given XML file is empty. Please provide the correct name before saving the file.")
            self.videoname="NO-NAME"
        else:
            self.videoname=x.text        
        
        # Get all the actions
        actionnodes=vid.findall("action")
        if len(actionnodes)<1:
            tkMessageBox.showinfo(TITLE, "The given XML file does not contain any actions.")
        for a in actionnodes:
            # Add the classnr to the runid array. Grow if necessary
            anr=int(getAtt(a,"nr"))
            aclass=int(getAtt(a,"class"))
            # print "-----",anr,aclass
            # print len(self.runids)
            if len(self.runids)<anr:
                # print "Growing runid:"
                self.runids += [None]*(anr-len(self.runids))            
            self.runids[anr-1]=aclass
            # print "size of runid array:", len(self.runids), "array:", self.runids    
            
            # Get all the bounding boxes for this action
            bbs=a.findall("bbox")
            if len(bbs)<1:
                tkMessageBox.showinfo(TITLE, "No <bbox> tags found for an action in the input XML file!")
                sys.exit(1)
            for bb in bbs:
                
                # Add the bounding box to the frames() list
                bfnr=int(getAtt(bb,"framenr"))
                bx=int(getAtt(bb,"x"))
                by=int(getAtt(bb,"y"))
                bw=int(getAtt(bb,"width"))
                bh=int(getAtt(bb,"height"))
                try:
                    self.addRect(bx,by,bx+bw-1,by+bh-1,anr,bfnr-1)
                except:
                    print "*** ERROR ***"
                    print "The XML file contains rectangles in frame numbers which are outside of the video"
                    print "(frame number too large). Please check whether the XML file really fits to these"
                    print "frames."
                    sys.exit(1)

# ***************************************************************************
# GUI
# The state variable self.state can take one of the following values:
# ul 	we are currently moving the upper left corner
# ur 	we are currently moving the upper right corner
# ll 	we are currently moving the lower left corner
# lr 	we are currently moving the lower right corner
# c 	we are currently moving the window
# d	we are currently drawing a new rectangle
# i	we are currently choosing the running id
# "" 	(empty) no current action
# ***************************************************************************

class Example(Frame):

    def __init__(self, parent, aCurPath):
        Frame.__init__(self, parent)            
        self.parent = parent     
        self.curPath = aCurPath
        self.ct = AAControler();        
        self.initUI()            
    
    # Interface startup: create all widgets and create key and mouse event
    # bindings
    def initUI(self):      
        self.parent.title(TITLE+" (frame nr.1 of "+str(len(self.ct.filenames))+")")        
        self.pack(fill=BOTH, expand=1)                
        self.img = self.ct.curFrame()
        self.curFrame = ImageTk.PhotoImage(self.img)
        # The grid layout
        self.canvas = Canvas(self.parent, width=self.img.size[0], 
            height=self.img.size[1])
        self.canvas.create_image(0, 0, anchor=NW, image=self.curFrame)                                
        #self.canvas.pack(side=LEFT)        
        self.runidbox = Listbox(self.parent)
        #self.runidbox.pack(side=RIGHT, fill=Y)                    
        self.quitbutton = Button(self.parent, text="QUIT")
        self.fnEntry = Entry(self.parent)
        #self.savebutton.pack()        
        self.grid(sticky=W+E+N+S)
        self.canvas.grid(row=0,column=0,rowspan=4)
        self.runidbox.grid(row=0,column=1,sticky=N+S)
        self.fnEntry.grid(row=1,column=1)
        self.quitbutton.grid(row=3,column=1)
        
        # Bindings - for both versions               
        self.canvas.bind ("<Key-Left>", self.prevFrame)
        self.canvas.bind ("<Key-BackSpace>", self.prevFrame)
        self.canvas.bind ("<Key-Right>", self.nextFrame)        
        self.canvas.bind ("<Next>", self.nextFrameFar)
        self.canvas.bind ("<Prior>", self.prevFrameFar)   # the space key
        self.quitbutton.bind("<Button-1>", self.quit) 
        self.canvas.bind ("<Motion>", self.mouseMove)
 
        
        # Variable inits
        self.state=""
        self.mousex = 1
        self.mousey = 1
        self.runIdProposed4NewRect=1
        self.displayAnno()
        self.displayRunIds()
        self.fnEntry.delete(0, END)
        self.fnEntry.insert(0, self.ct.videoname)
        self.isModified=False
        self.canvas.focus_force()
        
    def checkValidity(self):
        msg=self.ct.checkValidity()
        if len(self.fnEntry.get())<1:
            msg=msg+"The video name is empty.\n"
        if len(msg)>0:
            tkMessageBox.showinfo(TITLE, "There are errors in the annotation:\n\n"+msg+"\nThe file has been saved. Please address the problem(s) and save again.")    
        
    def quit(self,event):        
        print "quit method"
        self.ct.videoname=self.fnEntry.get()

        ok=True
        if ok:
            self.parent.destroy()


    def updateAfterJump(self):
        self.curFrame = ImageTk.PhotoImage(self.img)
        self.displayAnno()        
        self.parent.title(TITLE+" (frame nr."+str(self.ct.curFrameNr+1)+" of "+str(len(self.ct.filenames))+")")  
        self.canvas.update()
        
    def prevFrame(self,event):
        self.img = self.ct.prevFrame()
        self.updateAfterJump()
        
    def prevFrameFar(self,event):
        self.img = self.ct.prevFrameFar()
        self.updateAfterJump()
                
    def nextFrame(self,event):
        self.img = self.ct.nextFrame(False, False)
        self.updateAfterJump()
        
    def nextFrameFar(self,event):
        self.img = self.ct.nextFrameFar()
        self.updateAfterJump()
                
    def nextFrameWProp(self,event):        
        self.img = self.ct.nextFrame(True, False)
        self.updateAfterJump()
        self.isModified=True
        
    def nextFrameWPropForced(self,event):        
        self.img = self.ct.nextFrame(True, True)
        self.updateAfterJump()
        self.isModified=True
        
    def mouseMove(self,event):
        self.displayAnno()
        self.mousex = event.x
        self.mousey = event.y
        
        # Put the focus on the canvas, else the other widgets 
        # keep all keyboard events once they were selected.
        self.canvas.focus_force()


    # Draw the image and the current annotation
    def displayAnno(self):
        self.canvas.create_image(0, 0, anchor=NW, image=self.curFrame)	
        
        # We are currently in an operation, so do not search the nearest rectangle. It is
        # the one blocked at the beginning of the operation
        if self.state in ("ul","ur","ll","lr","c","d","i"):
            sempos = self.curSempos
        
        # Search for the nearest rectangle:
        # which rectangle is the nearest one to the mouse cursor, and what is
        # its relative position (corners, center, general position)?
        else:	
            sempos = self.ct.getSemMousePos(self.mousex,self.mousey)	
                
        # Draw all rectangles
        for (i,r) in enumerate(self.ct.getRects()):
            if i==sempos.index:
                curcol="blue"
            else:
                curcol="red"    
            self.canvas.create_rectangle(r.x1, r.y1, r.x2, r.y2, outline=curcol, 
                width=2)	    
            self.canvas.create_text(r.x1+15,r.y1+15, text=str(r.runid), 
                fill=curcol, font=("Helvectica", "20"))


    def displayRunIds(self):
        self.runidbox.delete(0, END) 
        x=self.ct.runids
        for i in range(len(x)):
            if x[i]<0:
                self.runidbox.insert(END, str(i+1)+" has no assigned class ")
            else:
                self.runidbox.insert(END, str(i+1)+" has class "+str(x[i])+" ["+classnames[x[i]-1]+"]")
               

def onexit():
    print "qqqq"
    ex.quit(None)
        
        
trackingLib = None
    
def main():
    curPath=sys.path[0]

    root = Tk()
    root.protocol("WM_DELETE_WINDOW", onexit) 
    global ex 
    ex = Example(root, curPath)
    root.mainloop()  

if __name__ == '__main__':
    main()  
